/*

i seni vanno da -1 a +1
comportarsi di conseguenza.

- Optimize Dots
- Dithering (cambiando le mask)
- Cel Shading (se il lato del triangolo confina con un triangolo cullato, allora tracciarlo)

*/

//#define _NeedMath_		// Do we need the mathlib?

#include <Pilot.h>
#include <SysEvtMgr.h>
#include <M68328Hwr.h>
#include <window.h>
#include <scrdriver.h>
#include <MemoryMgr.h>
#ifdef _NeedMath_
	#include <mathlib.h>
#endif
#include "types.h"

#include "RastyPalm_res.h"

// Sprite 4 colori.
// La struttura va bene sia se lo sprite e' packed (shift non precalcolati) sia se 'unpacked' (4 shift precalcolati)
// Cio' che cambia e' la routine di disegno.
typedef struct {
	short x, y;	// x and y positions on screen
	ushort frame;	// Frame corrente
	ushort sizex, sizey, framesize, bytesx;	// dimensioni X e Y (in pixel) e dimensione totale in bytes dello sprite (singolo frame), da precalcolare. Utile per saltare da un frame all'altro... bytesx e' (xsize/4)+1
	ushort frames;	// Numero di frames di animazione (minimo 1)
	ushort datax, datay; 	// Posizione iniziale dello sprite (x e y) all'interno di *bmp
	uchar *bmp;	// Actual sprite data. Precalcolare ogni elemento con lo stesso sprite shiftato.
	uchar *mask;	// Maschera con i pixel tutti a 11 per farci il not con la destinazione
} SpriteStruct4;

#include "spr_sprite1.h"

#define MPPerFace 3		// Max punti per faccia

typedef struct {
	short xsize, ysize;	// size x e y in pixels
	short bytesx;	// size X in bytes
	char colors;	// Number of colors
} ScreenStruct;

typedef struct {
	double x, y, z;
} Point3D;

typedef struct {
	long x, y, z;
} Point3Dl;

typedef struct {
	short x, y, z;
} Point3Ds;

typedef struct {
	short x, y;
} Point2D;

typedef struct {
	ulong face[MPPerFace];      // Punti faccia
	//ulong ppf;        // Punti Per Faccia
	uchar fcol;       // Colore faccia
	//uchar fattr;      // Attributi faccia
	//long zmed;       // Z media del poligono (Inutilizzata)
	uchar vis;        // 0=Faccia nascosta 1=Faccia visibile (x il Culling)
	//ulong mat;        // Materiale faccia
	//float nt;         // Intensita' della luce
} FaceStruct;

typedef struct {
	ulong NP;    // Numero punti nell'oggetto
	Point3Dl *p;  // Punti 3D originali         -- Non far mai puntare niente direttamente
	Point3Dl *r;  // Punti 3D ruotati           -- ai membri di questa struttura, perche'
	Point2D *s;  // Punti 3D (2D) trasformati   -- sono soggetti a realloc() !!!
	long XTrot, YTrot, ZTrot;     // Rotazioni totali (incrementano)
	ulong NF;    // Numero facce
	FaceStruct *f;    // Faccia[NF]

	long *zmed;
	ushort *zord;
	//ZetaMed *zm;  // Array di [NF] elementi per l'ordinamento dei poligoni
	ushort zmax;   // Puntatore all'ultimo elemento del'array di cui sopra

	long tx, ty, tz; // Nel lettore 3DS servono per la translation matrix
} Obj3D;

//#include "o3d_cubo.h"
//#include "o3d_aggr.h"
//#include "o3d_duck.h"
//#include "o3d_face.h"
//#include "o3d_tutwire.h"
#include "o3d_toro.h"

Obj3D *Obj;

// ---------- PROVA Punto 3D ------------
Point3D punto1, punto2;
Point3Dl punto1l, punto2l;
Point2D punto2d;
Point3D rot, rotinc;
Point3Dl rotl, rotincl;
short xxxx = 0;
// --------------------------------------

// ---------- Variabili temporanee ------------
char buffnum;	// Numero del buffer video attivo (0 o 1)
// --------------------------------------------

// ---------- 3D Stuff ------------
Point3Dl trans;		// Fixed 3D to 2D translation: x e y = 80, z = lontananza x la precisione
long Dfoc = 7;		// Perspective distortion
char o3d_playing;	// Animare o no?
// --------------------------------

// ---------- Screen definition -----------
ScreenStruct appscreen;
// ----------------------------------------

//short zoomfactor = 256;
short zoomfactor = 1024;

volatile HwrDBallType *_dragon = (HwrDBallType *)0xFFFFF000;

//unsigned char *VidMem = (UBYTE *)(0x9800);  // Pointer alla memoria video dello sfondo

unsigned char framebuffers[2][80*160];
unsigned char *curframebuffp = framebuffers[0];

UInt16 FormMain = 1000;
FormPtr FormMainID;

WinHandle WindowMain;
BitmapType *WindowBitmap;
unsigned char *VidMem;		// Pagina video reale
unsigned char VidBuf2[20*160];		// Screen buffer in memoria a 2 colori
unsigned char VidBuf4[40*160];		// Screen buffer in memoria a 4 colori
WinPtr winP;
static MenuBarPtr MainMenu;

unsigned short PolyOffset;
UInt32 Ticks;
UInt32 TicksDelay;

#ifdef _NeedMath_
long seno[627*2], coseno[627*2];
#else
#include "sintab.h"
#endif

short PreY[160];	// Y precalcolate * 20
short PreY4[160];	// Y precalcolate * 40
short BufY[160];	// Per fillare
unsigned char FillMaskStart[8] = { 255, 127, 63, 31, 15, 7, 3, 1 };
unsigned char FillMaskEnd[8] = { 128, 192, 224, 240, 248, 252, 254, 255 };

// FillMask[colore(0-3)][posizione colore nel byte]
unsigned char FillMaskStart4[4][4] = {	{ 0, 0, 0, 0},	// col. 00
										{ 85, 21, 5, 1},	// col. 01
										{ 170, 42, 10, 2},	// col. 10
										{ 255, 63, 15, 3} };	// col. 11

unsigned char FillMaskEnd4[4][4] = {	{ 0, 0, 0, 0},	// col. 00
										{ 64, 80, 84, 85},	// col. 01
										{ 128, 160, 168, 170},	// col. 10
										{ 192, 240, 252, 255} };	// col. 11

unsigned char SpriteMaskStart4[4] = {255, 63, 15, 3};	// 1111 0111 0011 0001
unsigned char SpriteMaskEnd4[4] = {192, 240, 252, 255};	// 1000 1100 1110 1111

// Quando si fa l'OR di un byte per disegnare un pixel a 4 colori,
// prima di tutto fare l'AND del colore "not 3" alla posizione corrispondente
// sul byte di destinazione, poi fare l'OR con la destinazione cosi' ottenuta.
//
// es.: colore 2  00 00 00 10 da plottare su
//      dest      11 00 00 01
//
//      not col.3 11 11 11 00 and
//      dest      11 00 00 01 =
//                11 00 00 00 (nuovo valore di dest)
//
//      colore 2  00 00 00 10 or
//      dest      11 00 00 00 =
//                11 00 00 10

// Setta l'indirizzo iniziale della memoria video
void LCD_SetAddress( void *memp )
{
	//_dragon->lcdStartAddr = (DWord)memp;
	*(DWord *)0xFFFFFA00 = (DWord)memp;
}

void LCD_GrayScale( char onoff )
{
	if ( onoff )
	{
		_dragon->lcdPageWidth = 20;
		_dragon->lcdPanelControl |= 0x01;
		_dragon->lcdLastBufferAddr = 20;
		_dragon->lcdFrameRate = 0xB9;
	}
	else
	{
		_dragon->lcdPageWidth = 10;
		_dragon->lcdPanelControl &= 0xFE;
		_dragon->lcdLastBufferAddr = 10;
	}

}

static void putpixel2( short x, short y )
{
	unsigned char	*memp, bitmsk;

	memp = VidMem + (x >> 3) + PreY[y];
	bitmsk = 1 << (7 - (x & 7));
	*memp |= bitmsk;
}

static void putpixel4( short x, short y, unsigned char col )
{
	unsigned char	*memp, bitmsk;

	memp = VidMem + (x >> 2) + PreY4[y];

	bitmsk = col << ((3 - (x & 3))<<1);
	*memp |= bitmsk;
}

static void ClearScreen2(uchar *p)
{
	// 3200 bytes
	memset(p, 0, 3200);
}

static void ClearScreen4(uchar *p)
{
	// 6400 bytes
	//memset(p, 0, 6400);
	ushort i;
	for (i=0; i<6400; i++) p[i]=0;
}

void inline FillLine2(unsigned short Py, unsigned short Ex)
{
  unsigned short Tx, Sx;
  unsigned short Sx8, Ex8, SxByte, ExByte, Ypos;
  unsigned short Sx8Y1, Ex8Y;
  unsigned char i;
  unsigned short c;

  Sx = BufY[Py];

  if (Sx>Ex) {  // Sx deve essere < Ex
    Tx = Sx;
    Sx = Ex;
    Ex = Tx;
  }

  Ypos = PreY[Py];

  Sx8 = Sx >> 3;
  SxByte = Sx8 + Ypos;
  Ex8 = Ex >> 3;

  //if ((Ex-Sx) < 8) {
  if (Sx8 == Ex8) {
  	//for (i=Sx; i<=Ex; i++) putpixel2(i, Py);
  	//for (c=Sx; c<=Ex; c++) VidMem[(c>>3)+Ypos] |= (1 << (7 - (c & 7)));
  	// Metodo alternativo: cioccare se Sx8 == Ex8.
  	// Se si, fare la negazione dell'or esclusivo della maschera iniziale e finale...!
  	VidMem[SxByte] = VidMem[SxByte] | ~(FillMaskStart[Sx&7] ^ FillMaskEnd[Ex&7]);
  } else {
  // ...
  // Sx/8 + 1 = Byte di start in VidMem
  // Sx/8 = Byte in VidMem da fillare con FillMaskStart[ Sx & 7 ]
  // Ex/8 - 1 = Byte di end in VidMem
  // Ex/8 = Byte in VidMem da fillare con FillMaskEnd[ Ex & 7 ]
  // Fillare bytes da Sx/8+1 ad Ex/8-1

  //Sx8 = Sx >> 3;
  //SxByte = Sx8 + Ypos;
  //Ex8 = Ex >> 3;
  ExByte = Ex8 + Ypos;  
  VidMem[SxByte] = VidMem[SxByte] | FillMaskStart[ Sx & 7 ];
  VidMem[ExByte] = VidMem[ExByte] | FillMaskEnd[ Ex & 7 ];
  //for (i=Sx8+1; i<Ex8; i++) VidMem[i+Ypos]=255;
  Sx8Y1 = Sx8+1+Ypos;
  Ex8Y = Ex8+Ypos;
  for (c=Sx8Y1; c<Ex8Y; c++) VidMem[c]=255;

//  putpixel2(50, 50);  
  }
}

void inline FillLine4(unsigned short Py, unsigned short Ex, unsigned char col, uchar *Dest)
{
  unsigned short Tx, Sx;
  unsigned short Sx4, Ex4, SxByte, ExByte, Ypos;
  unsigned short Sx4Y1, Ex4Y;
  unsigned char i;
  unsigned short c;

  Sx = BufY[Py];

  if (Sx>Ex) {  // Sx deve essere < Ex
    Tx = Sx;
    Sx = Ex;
    Ex = Tx;
  }

  Ypos = PreY4[Py];


  Sx4 = Sx >> 2;
  SxByte = Sx4 + Ypos;
  Ex4 = Ex >> 2;
  //if ((Ex-Sx) < 8) {
  if (Sx4 == Ex4) {
	  //for (c=Sx; c<Ex; c++) putpixel4(c, Py, col);
	  
	  i = (FillMaskStart4[3][Sx&3] & FillMaskEnd4[3][Ex&3]);
	  Dest[SxByte] &= ~i;

	  Dest[SxByte] = Dest[SxByte] | (FillMaskStart4[col][Sx&3] & FillMaskEnd4[col][Ex&3]);

	  //Dest[SxByte] &= ~(FillMaskStart4[3][Sx&3]);
	  //Dest[SxByte] &= ~(FillMaskEnd4[3][Ex&3]);
	  //Dest[SxByte] &= ~i;
	  //Dest[SxByte] = Dest[SxByte] | i;
	  //Dest[SxByte] = i;
  } else {
  // ...
  // Sx/8 + 1 = Byte di start in VidMem
  // Sx/8 = Byte in VidMem da fillare con FillMaskStart[ Sx & 7 ]
  // Ex/8 - 1 = Byte di end in VidMem
  // Ex/8 = Byte in VidMem da fillare con FillMaskEnd[ Ex & 7 ]
  // Fillare bytes da Sx/8+1 ad Ex/8-1

	  ExByte = Ex4 + Ypos;  
	  Dest[SxByte] &= ~(FillMaskStart4[3][Sx&3]);
	  Dest[ExByte] &= ~(FillMaskEnd4[3][Ex&3]);
	  Dest[SxByte] = Dest[SxByte] | FillMaskStart4[col][ Sx & 3 ];
	  Dest[ExByte] = Dest[ExByte] | FillMaskEnd4[col][ Ex & 3 ];
	  Sx4Y1 = Sx4+1+Ypos;
	  Ex4Y = Ex4+Ypos;
	  for (c=Sx4Y1; c<Ex4Y; c++) Dest[c]=FillMaskStart4[col][0];
	
	  //for (c=Sx; c<Ex; c++) putpixel4(c, Py, col);
  }
  
  //for (c=Sx; c<Ex; c++) putpixel4(c, Py, col);
}

void FlatPoly(unsigned short Sx, unsigned short Sy, unsigned short Ex, unsigned short Ey, unsigned char col, unsigned char *Dest)
{
  long DeltaX;  // Puo' anche essere negativo...!
  unsigned short DifX;
  unsigned short DifY;
  unsigned short tmp;
  const unsigned short ScreenX = 160;
  long xx, tmpxx;
  unsigned short yy, yy3;
  unsigned short PixPos;

 if (Sy!=Ey) {
    if (Sy>Ey) {  // Sy deve essere < di Ey
      tmp = Sx;
      Sx = Ex;
      Ex = tmp;
      tmp = Sy;
      Sy = Ey;
      Ey = tmp;
    }
    DifX = Ex - Sx;
    DifY = Ey - Sy;

    DeltaX = DifX;
    DeltaX <<= 16;
    (DifY!=0) ? (DeltaX /= DifY) : (DeltaX=0);

    xx=Sx;
    xx <<= 16;

    //yy3=Sy * ScreenX;
    for (yy=Sy; yy<Ey; yy++) {
      //PixPos=((xx >> 16) );
      tmpxx = xx >> 16;
      PixPos = tmpxx;
      if (BufY[yy] == 9999)
        BufY[yy] = PixPos;
      else {
//        if (BufY[yy]!=PixPos) {   // Non filla se Start=End
          //Xfillc(BufY[yy]+(yy3),PixPos+yy3,col,Dest);
          FillLine2(yy,PixPos);
          BufY[yy] = 9999;
//        }
      }
      //yy3 += ScreenX;
      xx += DeltaX;
    }
  }

}

void FlatPoly4(unsigned short Sx, unsigned short Sy, unsigned short Ex, unsigned short Ey, unsigned char col, unsigned char *Dest)
{
  long DeltaX;  // Puo' anche essere negativo...!
  unsigned short DifX;
  unsigned short DifY;
  unsigned short tmp;
  const unsigned short ScreenX = 160;
  long xx, tmpxx;
  unsigned short yy, yy3;
  unsigned short PixPos;

 if (Sy!=Ey) {
    if (Sy>Ey) {  // Sy deve essere < di Ey
      tmp = Sx;
      Sx = Ex;
      Ex = tmp;
      tmp = Sy;
      Sy = Ey;
      Ey = tmp;
    }
    DifX = Ex - Sx;
    DifY = Ey - Sy;

    DeltaX = DifX;
    DeltaX <<= 16;
    (DifY!=0) ? (DeltaX /= DifY) : (DeltaX=0);

    xx=Sx;
    xx <<= 16;

    //yy3=Sy * ScreenX;
    for (yy=Sy; yy<Ey; yy++) {
      //PixPos=((xx >> 16) );
      tmpxx = xx >> 16;
      PixPos = tmpxx;
      if (BufY[yy] == 9999)
        BufY[yy] = PixPos;
      else {
//        if (BufY[yy]!=PixPos) {   // Non filla se Start=End
          //Xfillc(BufY[yy]+(yy3),PixPos+yy3,col,Dest);
          FillLine4(yy,PixPos,col, Dest);
          BufY[yy] = 9999;
//        }
      }
      //yy3 += ScreenX;
      xx += DeltaX;
    }
  }

}

static void PreCalcY()
{
	short i;
	for (i=0; i<160; i++) PreY[i] = i * 20;
	for (i=0; i<160; i++) PreY4[i] = i * 40;
	for (i=0; i<160; i++) BufY[i] = 9999;
}

static void FillBufY()
{
	short i;
	for (i=0; i<160; i++) BufY[i] = 9999;
}

// Modalita' 4 colori: 2 bits per pixel, ogni byte contiene 4 pixel.
// Lo schermo diventa lungo 40 bytes.
static void DrawGrayscale()
{
	unsigned short i;
	unsigned char col;;
/*	
	for (i=PreY[150], col=1; i<PreY[150]+40; i++) {
		VidMem[i] = col;
		col <<=1;
	}
*/	
	for (i=PreY[100], col=1; i<PreY[100]+(40*50); i+=40) {
		//VidMem[i] = col | 1;
		//col <<=1;
		VidMem[i] = col++;
	}

}

void o3d_ActivateObj(Obj3D *xobj)
{
  Obj = xobj;
}

static Boolean MainViewHandleEvent(EventPtr event)
{
	Boolean		handled = false;

	if (event->eType == ctlSelectEvent)
	{
   		//if (event->data.ctlEnter.controlID == 1005)
		//{
		//	SysReset();
		//	handled = true;
		//}
	}
	else if (event->eType == penDownEvent)
	{
		handled = FrmHandleEvent (FrmGetActiveForm (), event);
	}

	return(handled);
}

static void DrawSine()
{
	UInt16 i;
	
	for (i=0; i<100; i++) putpixel4(coseno[i], i, 1);
}
/*
// Trasforma da 3Dfloat a 2Dint
static void ObjTrans()
{
  ulong i;

  for (i=0; i<Obj->NP; i++) {
    if ( (Obj->r[i].z - Obj->tz) != 0) {
      Obj->s[i].x =  ( ((long)Obj->r[i].x << Dfoc) / (Obj->r[i].z-Obj->tz) ) + Obj->tx;
      Obj->s[i].y =  ( ((long)Obj->r[i].y << Dfoc) / (Obj->r[i].z-Obj->tz) ) + Obj->ty;
    } else {
      Obj->s[i].x = ( (long)Obj->r[i].x << Dfoc) + Obj->tx;
      Obj->s[i].y = ( (long)Obj->r[i].y << Dfoc) + Obj->ty;
    }
  }

}
*/

static void qsortpolys( long left, long right )
{
	long j;
	static long i;
	static long piv, x;
	static ushort y;

	if ( left >= right )    return;
	
	piv = Obj->zmed[left];
	//piv = Obj->zm[right>>1].zmed;
	
	i = left + 1;
	j = right;
	for(;;)
	{
		while ( Obj->zmed[i] < piv && i<right ) ++i;
		while ( Obj->zmed[j] > piv && j>left ) --j;
        if ( i < j )
        {
            x = Obj->zmed[j];
            Obj->zmed[j] = Obj->zmed[i];
            Obj->zmed[i] = x;
            y = Obj->zord[j];
            Obj->zord[j] = Obj->zord[i];
            Obj->zord[i] = y;
            ++i;
            --j;
        }
        else
                break;
	}
	
	x = Obj->zmed[j];
	Obj->zmed[j] = Obj->zmed[left];
	Obj->zmed[left] = x;
	y = Obj->zord[j];
	Obj->zord[j] = Obj->zord[left];
	Obj->zord[left] = y;
	
	qsortpolys( left, j-1 );
	qsortpolys( j+1,  right );
}

//qsortpolys( 0, Obj->NF - 1 );

static void DotTrans(Point3D *r, Point2D *s)
{
	short tx = 80;
	short ty = 80;
	short tz = 60;
	short Dfoc = 7;
	Point3D ss;

	if ( (r->z - tz) != 0) {
		ss.x = ( ((long)r->x << Dfoc) / (r->z - tz)) + tx;
		ss.y = ( ((long)r->y << Dfoc) / (r->z - tz)) + ty;
	} else {
		ss.x = ( (long)r->x << Dfoc) + tx;
		ss.y = ( (long)r->y << Dfoc) + ty;
	}

	s->x = (short)ss.x;
	s->y = (short)ss.y;

	if (s->x >= 160 || s->x < 0) s->x = 0;
	if (s->y >= 160 || s->y < 0) s->y = 0;
}

static void DotTransl(Point3Dl *r, Point2D *s)
{
	long tx = 80;
	long ty = 80;
	long tz = 360;
	Point3Dl ss;

	if ( (r->z - trans.z) != 0) {
		ss.x = ( (r->x << Dfoc) / (r->z - trans.z)) + trans.x;
		ss.y = ( (r->y << Dfoc) / (r->z - trans.z)) + trans.y;
	} else {
		ss.x = (r->x << Dfoc) + tx;
		ss.y = (r->y << Dfoc) + ty;
	}

	s->x = ss.x;
	s->y = ss.y;

	if (s->x >= 160 || s->x < 0) s->x = 0;
	if (s->y >= 160 || s->y < 0) s->y = 0;
}


static void RotXYZ(Point3D *dot, Point3D *dotrot, double rotx, double roty, double rotz)
{

  double ox, oy, oz, ox1, oy1, oz1;
  double sa, csa;

  // Rotazione X
  sa = sin(rotx);
  csa = cos(rotx);
  ox = (dot->x*1);
  oy = (dot->y*1) * csa + (dot->z*1) * -sa;
  oz = (dot->y*1) * sa + (dot->z*1) * csa;

  // Rotazione Y
  sa = sin(roty);
  csa = cos(roty);
  ox1 = ox * csa + oz * -sa;
  oy1 = oy;
  oz1 = ox * sa + oz * csa;

  // Rotazione Z
  sa = sin(rotz);
  csa = cos(rotz);
  dotrot->x = ox1 * csa + oy1 * -sa;
  dotrot->y = ox1 * sa + oy1 * csa;
  dotrot->z = oz1;
}

static void RotXYZ2(Point3Dl *dot, Point3Dl *dotrot, long rotx, long roty, long rotz)
{

	long ox, oy, oz, ox1, oy1, oz1;
	long sa, csa;

	rotx = (zoomfactor-1) & rotx;
	roty = (zoomfactor-1) & roty;
	rotz = (zoomfactor-1) & rotz;

	// Rotazione X
	sa = seno[rotx];
	csa = coseno[rotx];
	ox = (dot->x);
	oy = (((dot->y) * csa + (dot->z) * -sa)) >> 10;
	oz = (((dot->y) * sa + (dot->z) * csa)) >> 10;

	// Rotazione Y
	sa = seno[roty];
	csa = coseno[roty];
	ox1 = ((ox * csa + oz * -sa)) >> 10;
	oy1 = oy;
	oz1 = ((ox * sa + oz * csa)) >> 10;

	// Rotazione Z
	sa = seno[rotz];
	csa = coseno[rotz];
	dotrot->x = ((ox1 * csa + oy1 * -sa)) >> 10;
	dotrot->y = ((ox1 * sa + oy1 * csa)) >> 10;
	dotrot->z = oz1;
}

static void o3d_Rotate(Point3Dl *rr)
{
	ulong dots;
	for (dots=0; dots<Obj->NP; dots++) {
		RotXYZ2(&Obj->p[dots], &Obj->r[dots], rr->x, rr->y, rr->z);
		DotTransl(&Obj->r[dots], &Obj->s[dots]);
		//putpixel4(Obj->s[dots].x, Obj->s[dots].y, 2);
	}
	//RotXYZ2(&punto1l, &punto2l, rotl.x, rotl.y, rotl.z);
}

static void o3d_Trans()
{
}

// Precalcola le facce nascoste
void PreCulling()
{
	Point3Dl P1, P2, P3;
	long d1, d2, d3, d4;
	ulong i;
	for (i=0; i<Obj->NF; i++) {
		P1.x = Obj->s[ Obj->f[i].face[0] ].x;
		P1.y = Obj->s[ Obj->f[i].face[0] ].y;
		P2.x = Obj->s[ Obj->f[i].face[1] ].x;
		P2.y = Obj->s[ Obj->f[i].face[1] ].y;
		P3.x = Obj->s[ Obj->f[i].face[2] ].x;
		P3.y = Obj->s[ Obj->f[i].face[2] ].y;
		d1=P1.x - P2.x;
		d3=P1.y - P2.y;
		
		d4=P2.y - P3.y;
		d2=P2.x - P3.x;
		
		d4=d1*d4;
		d2=d3*d2;
		
		if (d2<=d4) Obj->f[i].vis = 1; // Faccia visibile
			else Obj->f[i].vis = 0;
	}
}

/*

	Obj->zmax = 0;
	loop 'i' per tutti polys {
		Culling poly
		il poly e' stato cullato?
		no (e' visibile) {
			Obj->zm[ Obj->zmax ].ord = i;
			Obj->zm[ Obj->zmax++ ].zmed = zmed del poligono;
		}
	}
	
	qsort(0, Obj->zmax); sort dei polys visibili
	loop 'i' da 0 a Obj->zmax (quindi solo i visibili) {
		crea poligoni con ordine Obj->zm[i].ord;
		(NDX = Obj->zm[i].ord;)
		FlatPoly4(Obj->s[ Obj->f[NDX].face[0] ].x, Obj->s[ Obj->f[NDX].face[0] ].y, Obj->s[ Obj->f[NDX].face[1] ].x, Obj->s[ Obj->f[NDX].face[1] ].y, Obj->f[NDX].fcol, VidBuf4);
		
	}

*/

static void o3d_RenderFlat4()
{
	ulong faces;
	ushort i;
	long *vidmemp = (long *)VidMem;
	long *videbuf4p = (long *)VidBuf4;
	ushort NDX;	
	long zmed;

	// Per il culling
	Point3Dl P1, P2, P3;
	long d1, d2, d3, d4;

	Obj->zmax = 0;
	for (i=0; i<Obj->NF; i++) {
		// Culling
		P1.x = Obj->s[ Obj->f[i].face[0] ].x;
		P1.y = Obj->s[ Obj->f[i].face[0] ].y;
		P2.x = Obj->s[ Obj->f[i].face[1] ].x;
		P2.y = Obj->s[ Obj->f[i].face[1] ].y;
		P3.x = Obj->s[ Obj->f[i].face[2] ].x;
		P3.y = Obj->s[ Obj->f[i].face[2] ].y;
		d1=P1.x - P2.x;
		d3=P1.y - P2.y;
		
		d4=P2.y - P3.y;
		d2=P2.x - P3.x;
		
		d4=d1*d4;
		d2=d3*d2;
		
		//	d2 = 1;
		//	d4 = 2;
		
		if (d2<=d4) {
			Obj->f[i].vis = 1; // Faccia visibile
			Obj->zord[ Obj->zmax ] = i;	// Essendo visibile la aggiunge a quelle da sortare e displayare
			//zmed = Obj->r[ Obj->f[i].face[0] ].z + Obj->r[ Obj->f[i].face[1] ].z + Obj->r[ Obj->f[i].face[2] ].z;
			zmed = Obj->r[ Obj->f[i].face[0] ].z;
			if (Obj->r[ Obj->f[i].face[1] ].z > zmed) zmed = Obj->r[ Obj->f[i].face[1] ].z;
			if (Obj->r[ Obj->f[i].face[2] ].z > zmed) zmed = Obj->r[ Obj->f[i].face[2] ].z;
			//zmed /= 3;
			Obj->zmed[ Obj->zmax++ ] = zmed; // zmed del poligono
		} else {
			Obj->f[i].vis = 0;
		}
	}

	qsortpolys( 0, Obj->zmax - 1);

	//for (i=0; i<6400; i++) VidBuf4[i] = 0;
	for (i=6400/4; i > 0; --i) *videbuf4p++ = 0;
	//i = 6400/4; do { *videbuf4p++ = 0; } while (--i);
	videbuf4p = (long *)VidBuf4;

	for (i=0; i<Obj->zmax; i++) {
		NDX = Obj->zord[i];
		FlatPoly4(Obj->s[ Obj->f[NDX].face[0] ].x, Obj->s[ Obj->f[NDX].face[0] ].y, Obj->s[ Obj->f[NDX].face[1] ].x, Obj->s[ Obj->f[NDX].face[1] ].y, Obj->f[NDX].fcol, VidBuf4);
		FlatPoly4(Obj->s[ Obj->f[NDX].face[1] ].x, Obj->s[ Obj->f[NDX].face[1] ].y, Obj->s[ Obj->f[NDX].face[2] ].x, Obj->s[ Obj->f[NDX].face[2] ].y, Obj->f[NDX].fcol, VidBuf4);
		FlatPoly4(Obj->s[ Obj->f[NDX].face[2] ].x, Obj->s[ Obj->f[NDX].face[2] ].y, Obj->s[ Obj->f[NDX].face[0] ].x, Obj->s[ Obj->f[NDX].face[0] ].y, Obj->f[NDX].fcol, VidBuf4);
	}
	
	//for (i=0; i<6400; i++) VidMem[i] = VidBuf4[i];
	for (i=6400/4; i > 0; --i) *vidmemp++ = *videbuf4p++;
	//i = 6400/4; do { *vidmemp++ = *videbuf4p++; } while (--i);

}

static void o3d_xRenderFlat4()
{
	ulong faces;
	ushort i;
	long *vidmemp = (long *)VidMem;
	long *videbuf4p = (long *)VidBuf4;

	PreCulling();
	// Sortare e poi loopare solo per le le facce visibili sortate
	//for (i=0; i<6400; i++) VidBuf4[i] = 0;
	for (i=6400/4; i > 0; --i) *videbuf4p++ = 0;
	//i = 6400/4; do { *videbuf4p++ = 0; } while (--i);
	videbuf4p = (long *)VidBuf4;
	for (faces=0; faces<Obj->NF; faces++) {
		if (Obj->f[faces].vis) {
			FlatPoly4(Obj->s[ Obj->f[faces].face[0] ].x, Obj->s[ Obj->f[faces].face[0] ].y, Obj->s[ Obj->f[faces].face[1] ].x, Obj->s[ Obj->f[faces].face[1] ].y, Obj->f[faces].fcol, VidBuf4);
			FlatPoly4(Obj->s[ Obj->f[faces].face[1] ].x, Obj->s[ Obj->f[faces].face[1] ].y, Obj->s[ Obj->f[faces].face[2] ].x, Obj->s[ Obj->f[faces].face[2] ].y, Obj->f[faces].fcol, VidBuf4);
			FlatPoly4(Obj->s[ Obj->f[faces].face[2] ].x, Obj->s[ Obj->f[faces].face[2] ].y, Obj->s[ Obj->f[faces].face[0] ].x, Obj->s[ Obj->f[faces].face[0] ].y, Obj->f[faces].fcol, VidBuf4);
		}
	}
	//for (i=0; i<6400; i++) VidMem[i] = VidBuf4[i];
	for (i=6400/4; i > 0; --i) *vidmemp++ = *videbuf4p++;
	//i = 6400/4; do { *vidmemp++ = *videbuf4p++; } while (--i);
}

// no!
static void vid_SwitchBuffer4()
{
	ushort i;
	if (buffnum == 0) {
		buffnum = 1;
		for (i=0; i<6400; i++) VidMem[i]=0;
		LCD_SetAddress(VidMem + 6400);
		//MemSet(VidMem-6400, 0, 6400);
	} else {
		buffnum = 0;
		for (i=0; i<6400; i++) VidMem[i]=0;
		LCD_SetAddress(VidMem - 6400);
		//MemSet(VidMem+6400, 0, 6400);
	}
}

static Boolean MainFormDoMenuCommand (Word command)
{
	Boolean		handled = false;

	//o3d_playing = 0;
	MenuEraseStatus (0);
	switch (command) {
		case HelpAbout:
			FrmAlert(1000);
			o3d_playing = 1;
			handled = true;
		break;
	}
}

// Fare anche putbobpos4() dove gli si passano posizione x e y.
void putbob4(SpriteStruct4 *spr, uchar *dest)
{
	uchar *spritebmp, *spritemask;
	short xstart, xshift, ystart, ydif, xdif, inc, bytesx, xstartspr;
	short mod, xx, yy;
	uchar vis;
	ulong frame;

	vis = 1;	// sprite visibile
	
	bytesx = spr->bytesx;
	if (spr->x+spr->sizex >= appscreen.xsize) {	// Clipping right
		xstart = spr->x;
		if (xstart < (appscreen.xsize)) {
			xdif = (xstart + spr->sizex) - appscreen.xsize;
			xdif >>= 2;
			xdif++;		// lo sprite in realta' occupa 4 pixel in piu' sulle x delle dimensioni indicate da sizex
			bytesx -= xdif;
			xstartspr = xdif;
		} else {
			vis = 0;
		}
	}
	if (spr->x < 0) {	// Clipping left
		xstart = spr->x * -1;
		if (xstart <= spr->sizex) {	// se lo sprite non esce fuori del tutto...
			xshift = (3 & xstart);
			xstart >>= 2;
			xshift = 3 - xshift;
			spritebmp = spr->bmp + spr->framesize * xshift;
			spritemask = spr->mask + spr->framesize * xshift;
			xstartspr = xstart;	// posizione iniziale all'interno dello sprite
			spritebmp += xstart;
			spritemask += xstart;
			bytesx -= xstart;
			xstart = 0;
		} else {
			vis = 0;
		}
	}
	if (spr->x >= 0) {	// non e' clippato a sinistra
		xstart = spr->x;
		xstart >>= 2;
		xshift = 3 & spr->x;
		spritebmp = spr->bmp + spr->framesize * xshift;
		spritemask = spr->mask + spr->framesize * xshift;
	}
	mod = 40 - bytesx;
	dest += xstart;		// Start x in bytes sulla destinazione
	ystart = spr->y;
	yy = spr->sizey;

	if (ystart < 0) {	// Clipping top
		ydif = ystart * -1;	// Di quanto esce fuori lo sprite?
		if (ydif < yy) {	// Se lo sprite non esce fuori tutto...
			inc = spr->bytesx * ydif;
			spritebmp += inc;	// avanza il puntatore sullo sprite
			spritemask += inc;
			yy -= ydif;		// rimpicciolisci il size Y
			ystart = 0;
		} else vis = 0;	// Se lo sprite esce fuori tutto, invece, non disegnarlo.
	}
	if (ystart < appscreen.ysize) {	// Clipping bottom
		if (ystart+yy >= appscreen.ysize) {
			yy -= (ystart+yy) - appscreen.ysize;
		}
	} else vis = 0;

	dest += (PreY4[ystart]);

	// Disegna lo sprite
	if (vis) {		// Se lo sprite e' visibile...
		frame = spr->frame * spr->framesize;	// Punto al frame corretto
		frame <<= 2;	// frame corretto = frame attuale * size * 4
		spritebmp += frame;
		spritemask += frame;
 		if (bytesx != spr->bytesx)	// Se e' clippato sulle x...
		while (yy--) {
			xx = bytesx;	// Numero di bytes da plottare sulla linea
			while (xx--) {
				*dest &= ~*spritemask++;
				*dest++ |= *spritebmp++;
			}
			dest += mod;
			spritebmp += xstartspr;
			spritemask += xstartspr;
		}
		else
		if (spr->x >= 0 && spr->x+spr->sizex < appscreen.xsize)	// se non e' clippato sulle x...
		while (yy--) {
			xx = bytesx;	// Numero di bytes da plottare sulla linea
			while (xx--) {
				*dest &= ~*spritemask++;
				*dest++ |= *spritebmp++;
			}
			dest += mod;
		};
	}
	
}

static void drawsprites()
{
	/*
	uchar sizex;
	ushort sizey;
	uchar *spritebmp;
	spritebmp = spr_sprite1.bmp;
	
	for (sizey=0; sizey<spr_sprite1.sizey; sizey++) {
		for (sizex=0; sizex<spr_sprite1.bytesx; sizex++) {
			VidMem[(sizex)+PreY4[sizey]] = *spritebmp++;
		}
	}
	*/
	putbob4(&spr_sprite1, VidMem);
	spr_sprite1.x++;
	//spr_sprite1.y--;

}

static Boolean ApplicationHandleEvent(EventPtr event)
{
	Boolean	handled = false;

	FormPtr	frm;
	Int		formId;

	unsigned short i, c;
	
	//WinDrawPixel(50, 50);
	//WinPaintPixel(50, 50);

	/*VidMem[10] = 1;
	VidMem[20] = 1;
	VidMem[100] = 1;
	VidMem[150] = 1;
	VidMem[200] = 1;*/
	
	//for (i=0; i<200; i++) VidMem[i] = 10;
	
	FrmSetActiveForm(FormMainID);
//	FrmDrawForm(FormMainID);

	//for (i=0; i<256; i++) VidMem[i] = i;
	//for (i=0, c=0; i<100; i++, c+=2) putpixel2(c, 0);

	/*
	FlatPoly(50, 50, 100, 100, 1, VidMem);
	FlatPoly(100, 100, 30, 120, 1, VidMem);
	FlatPoly(50, 50, 30, 120, 1, VidMem);
	*/
	
	/*
	FlatPoly4(50+PolyOffset, 50, 100, 100, 1, VidMem);
	FlatPoly4(100, 100, 30, 120, 1, VidMem);
	FlatPoly4(50+PolyOffset, 50, 30, 120, 1, VidMem);
	*/

//	if (TimGetTicks() >= Ticks+TicksDelay) {
//		PolyOffset++;
//		Ticks = TimGetTicks();
//	}

	//DrawGrayscale();
	//DrawSine();
/*
	RotXYZ(&punto1, &punto2, rot.x, rot.y, rot.z);
	rot.x += rotinc.x;
	rot.y += rotinc.y;
	rot.z += rotinc.z;
	DotTrans (&punto2, &punto2d);
	putpixel4((short)punto2d.x, (short)punto2d.y, 2);
*/

/*
	RotXYZ2(&punto1l, &punto2l, rotl.x, rotl.y, rotl.z);
	rotl.x += rotincl.x;
	rotl.y += rotincl.y;
	rotl.z += rotincl.z;
	DotTransl(&punto2l, &punto2d);
	putpixel4(punto2d.x, punto2d.y, 2);
*/


	//vid_SwitchBuffer4();	
	//ClearScreen4(VidMem);
/*
	if (o3d_playing) {
		o3d_Rotate(&rotl);
		o3d_Trans();
		o3d_RenderFlat4();
		rotl.x += rotincl.x;
		rotl.y += rotincl.y;
		rotl.z += rotincl.z;
	}
*/
	
	//LCD_SetAddress(VidMem + 80);
	//LCD_SetAddress(VidMem);
	
/*
	char	statusStr[30] = {0};
	short	stringIndex = 0;

	if (event->eType == frmLoadEvent)
	{
		formId = event->data.frmLoad.formID;
		frm = FrmInitForm(formId);
		FrmSetActiveForm(frm);
		switch (formId)
		{
			case 1000:
				FrmSetEventHandler(frm, MainViewHandleEvent);
				break;
		}
		handled = true;
		if (SysLaunchConsole())
		{
			stringIndex++;
		}

		SysStringByIndex(1000, stringIndex, statusStr, sizeof(statusStr));
		FrmCopyLabel (frm, 1008, statusStr);
		FrmDrawForm(frm);
	}
	
	formId = event->data.frmLoad.formID;
	frm = FrmInitForm(formId);
	FrmSetActiveForm(frm);
	FrmDrawForm(frm);
*/

// Gestore eventi vero e proprio
	switch (event->eType) {
		case penDownEvent:
			o3d_playing = 0;
			FrmAlert(1000);
			o3d_playing = 1;
			handled = true;
		break;
		case menuEvent:
			handled = MainFormDoMenuCommand(event->data.menu.itemID);
		break;
		case frmCloseEvent:	// Chiudo il form
			if (event->data.frmClose.formID == 1000) {	// E' il form principale?
				//StopApplication();
			}
		break;
	}

	if (o3d_playing) {
		o3d_Rotate(&rotl);
		o3d_Trans();
		o3d_RenderFlat4();
		rotl.x += rotincl.x;
		rotl.y += rotincl.y;
		rotl.z += rotincl.z;
		drawsprites();
	}

	return handled;
}

static void IdleLoop()
{
	FlatPoly4(50+PolyOffset, 50, 100, 100, 1, VidMem);
	FlatPoly4(100, 100, 30, 120, 1, VidMem);
	FlatPoly4(50+PolyOffset, 50, 30, 120, 1, VidMem);

//	if (TimGetTicks() >= Ticks+TicksDelay) {
		PolyOffset++;
//		Ticks = TimGetTicks();
//	}
}

static void EventLoop(void)
{
	EventType	event;
	Word			error;
	
	do
		{
		

		// Get the next available event.
		//EvtGetEvent(&event, evtWaitForever);
		EvtGetEvent(&event, 5);
		//IdleLoop();
					
		// Give the system a chance to handle the event.
		if (! SysHandleEvent(&event))

			// P2. Give the menu bar a chance to update and handle the event.	
			if (! MenuHandleEvent(0, &event, &error)) {	// E' stato aperto un menu? se no, continua..
				// P3. Give the application a chance to handle the event.
				if (! ApplicationHandleEvent(&event))	// Detto anche MainFormHandleEvent, checka degli eventi
					// P3. Let the form object provide default handling of the event.
					FrmDispatchEvent(&event);
			} else {	// E' stato aperto un menu...
				o3d_playing = 0;
			}
		}
	while (event.eType != appStopEvent);
}

/*

Calcolo seno:

z = <zoom factor> (es 128, 256...)
inc = 1/z (incremento)
v = 0;
for (i=0; i<256; i++) {
	seno[i] = (long)(sin(v) * z);
	v += inc;
}

quindi come angoli di rotazione usare numeri da 0 a z (es. 256)
nel RotXYZ mettere "(z-1) and ANGOLO" (es. "255 and 260") per rientrare nell'array.
*/

#ifdef _NeedMath_
static void SaveSin(long *seno, long *coseno)
{
	long i;
	double x, x1=0,y1;
	UInt usecount;

	double inc, v;

	Err error;

	error = SysLibFind(MathLibName, &MathLibRef);
	if (error)
	    error = SysLibLoad(LibType, MathLibCreator, &MathLibRef);
	ErrFatalDisplayIf(error, "Can't find MathLib"); // Just an example; handle it gracefully
	error = MathLibOpen(MathLibRef, MathLibVersion);
	ErrFatalDisplayIf(error, "Can't open MathLib"); 
/*
  for (x=0, i=0; x<(627*2); x++, i++) {
    x1 = (256*(sin(x/256)));
    seno[i] = (long)x1;
    y1 = (256*(cos(x/256)));
    coseno[i] = (long)y1;
    //printf("*%d,%d*",seno[x], coseno[x]);
  }
*/
	error = MathLibClose(MathLibRef, &usecount);
	ErrFatalDisplayIf(error, "Can't close MathLib");
	if (usecount == 0)
	    SysLibRemove(MathLibRef); 

	//inc = 1;
	inc = 3.14159625*2;
	inc /= zoomfactor;
	//zoomfactor = 65536;
	//inc = 0.1;
	v = 0;	// -1
	//inc = 0.01;
	for (i=0; i<1024; i++, v++) {
		seno[i] = (long)(sin(v*((3.14159625*2)/zoomfactor))*zoomfactor);
		coseno[i] = (long)(cos(v*((3.14159625*2)/zoomfactor))*zoomfactor);
		//v += inc;
	}
  	
  	v = 0;	// <<- mettere breakpoint qui
/*
<kaztea> n_angles = 1024;
<kaztea> angles_to_radiants = 3.14159265 * 2 / n_angles
<kaztea> for (a=0; a < n_angles; ++a)
<kaztea>     sintable[a] = sin( a * angles_to_radiants );
<kaztea> poi.. per convertire da angoli 0-360 ad angoli 0-1024 usi una costante tipo (360/1024)
<kaztea> 0-1024, 0-360, 0-PI*2  ..sono semplicmente diversi modi di addressare gli stessi seni/coseni
<kaztea> solo che le funzioni sin() e cos() vogliono angoli in unita' 0-PI*2
<kaztea> mentre gli umani preferiscono usare 0-360.. e nei computer melgio usare qualcosa tipo 0-1024
<kaztea> poi puoi fare cosi'
<kaztea> int get_sin( int angle_0_360 )
<kaztea> {
<kaztea>   return sintable[ (a * 1024 / 360) & 1023 ];
<kaztea> }
*/
}
#endif

// if ( angle >= 90 && angle <= 180 ) return -table[angle-90];
// if ( angle >= 270 && angle <= 360 ) return table[360-angle];

static void InitApplication()
{
	PreCalcY();
	#ifdef _NeedMath_
	SaveSin(seno, coseno);
	#endif
	//MenuInit(1000);
	//MenuSetActiveMenuRscID(1000);
	appscreen.xsize = 160;
	appscreen.ysize = 160;
	appscreen.bytesx = 40;
	appscreen.colors = 4;
}


DWord PilotMain(Word cmd, Ptr cmdPBP, Word launchFlags)
{
	Err err;
	DWord newDefaultDepth; // = 1 or 2
	DWord newDepth; // = 1 or 2

	// ----------- PROVA Punto 3D -------------
	trans.x = 80;
	trans.y = 80;
	trans.z = 360;

	punto1.x = 0;
	punto1.y = 20;
	punto1.z = 20;
	rot.x = 0.0;
	rot.y = 0.0;
	rot.z = 0.0;
	rotinc.x = 0.0;
	rotinc.y = 0.2;
	rotinc.z = 0.0;

	//punto1l.x = 10;
	//punto1l.y = 10;
	//punto1l.z = -5;
	punto1l.x = 0;
	punto1l.y = 120;
	punto1l.z = 120;
	rotl.x = 0;
	rotl.y = 0;
	rotl.z = 0;
	//rotincl.x = 3;
	//rotincl.y = 5;
	//rotincl.z = 0;
	rotincl.x = 5;
	rotincl.y = 8;
	rotincl.z = 2;

	// ----------------------------------------
	//o3d_ActivateObj(&o3d_Cubo);
	//o3d_ActivateObj(&o3d_aggr);
	//o3d_ActivateObj(&o3d_duck);
	//o3d_ActivateObj(&o3d_face);
	//o3d_ActivateObj(&o3d_tutwire);
	o3d_ActivateObj(&o3d_toro);
	buffnum = 0;

	if (cmd == sysAppLaunchCmdNormalLaunch)
	{
		//InitApplication();
		//FrmGotoForm(1000);
		//curframebuffp = WinGetDisplayWindow()->displayAddr;
		FormMainID = FrmInitForm(1000);
		FrmSetActiveForm(FormMainID);
		FrmDrawForm(FormMainID);
		WindowMain = WinGetDrawWindow();
		InitApplication();
		o3d_playing = 1;
		//WindowBitmap = WinGetBitmap(WindowMain);
		//WinGetBitmap(WindowMain);
		//VidMem = BmpGetBits(WindowBitmap);
		
		winP = WinGetWindowPointer(WindowMain);
		//VidMem = ((WordPtr)winP->gDeviceP->baseAddr);
		VidMem = winP->displayAddrV20;

		TicksDelay = 25*20; // (mezzo secondo)
		Ticks = TimGetTicks();
		PolyOffset = 0;

		//LCD_GrayScale(1);
		// Let's be OS compatible!
		//newDefaultDepth = 2;
		//err = FtrSet(sysFtrCreator, sysFtrNumDisplayDepth, newDefaultDepth);
		// ScrDisplayMode funza alla prima botta, FtrSet no...!!
		newDepth = 2;
		err = ScrDisplayMode(scrDisplayModeSet, NULL, NULL,&newDepth, NULL);

		spr_sprite1.x = 1;
		spr_sprite1.frame = 1;
				
		EventLoop();

		newDefaultDepth = 1;
		err = FtrSet(sysFtrCreator, sysFtrNumDisplayDepth, newDefaultDepth);
	}

	return 0;
}

// UInt16 SysTicksPerSecond (void)
// UInt32 TimGetTicks (void)
// MemPtrFree(winH);

/*

//
// -------------------------------------------------------------------------
//   get_pivot - return the index of the selected pivot value
//

int get_pivot (int low, int hi) {

  // safety net, this should not happen
  if (low == hi) return(data[low]);

  // return the greater of the first two items in the range
  return( (data[low] > data[low+1]) ? low : (low+1) );
}

// -------------------------------------------------------------------------
//   swap - given two pointers to integers, swap their contents
//

void swap (int *a, int *b) {
  int temp = *a;
  *a = *b;
  *b = temp;
  num_swaps++;
}

// -------------------------------------------------------------------------
//   q_sort - Quicksort a data range
//

void q_sort (int low, int hi) {
  int pivot_index;                // index in the data set of the pivot
  int pivot_value;                // the value of the pivot element
  int left, right;

  // select the pivot element and remember its value
  pivot_index = get_pivot(low, hi);
  pivot_value = data[pivot_index];

  // do the partitioning
  left = low; right = hi;
  do {

    // move left to the right bypassing elements already on the correct side
    while ((left <= hi) && (data[left] < pivot_value)) {
      num_comps++;
      left++;
    }
    num_comps++;

    // move right to the left bypassing elements already on the correct side
    while ((right >= low) && (pivot_value < data[right])) {
      num_comps++;
      right--;
    }
    num_comps++;

    // 
    //  if the pointers are in the correct order then they are pointing to two
    //  items that are on the wrong side of the pivot value, swap them...
    //
    if (left <= right) {
      swap(&data[left], &data[right]);
      left++;
      right--;
    }
    
  } while (left <= right);

  // now recurse on both partitions as long as they are large enough
  if (low < right) q_sort(low, right);
  if (left < hi) q_sort(left, hi);
}

*/
